import pathlib

import bpy

from blenderline.entries.base import BaseEntry


class BackgroundEntry(BaseEntry):
    """Background entry."""

    def __init__(self, filepath: pathlib.Path, relative_frequency: float) -> None:
        """Create background entry.

        Args:
            filepath (pathlib.Path): absolute filepath to image asset.
            relative_frequency (float): relative frequency with which to sample.
        """
        # Save object attributes.
        self.filepath = filepath
        self.relative_frequency = relative_frequency

    def set(self, background_object: bpy.types.Object) -> None:
        """Apply background image to background object.

        Args:
            background_object (bpy.types.Object): object to apply background to.
        """
        # Get active material node tree.
        node_tree = background_object.active_material.node_tree
        nodes = node_tree.nodes

        # Clear all nodes.
        nodes.clear()

        # Add Texture Image node.
        node: bpy.types.ShaderNodeTexImage = nodes.new("ShaderNodeTexImage")
        texture_node = node
        texture_node.image = bpy.data.images.load(str(self.filepath))
        texture_node.location = (-300, 0)

        # Add Principled BSDF node.
        node: bpy.types.ShaderNodeBsdfPrincipled = nodes.new("ShaderNodeBsdfPrincipled")
        bsdf_node = node
        bsdf_node.location = (0, 0)

        # Add Material Output node.
        node: bpy.types.ShaderNodeOutputMaterial = nodes.new("ShaderNodeOutputMaterial")
        output_node = node
        output_node.location = (300, 0)

        # Link nodes.
        links = node_tree.links
        _ = links.new(
            input=texture_node.outputs["Color"],
            output=bsdf_node.inputs["Base Color"],
        )
        _ = links.new(
            input=texture_node.outputs["Alpha"],
            output=bsdf_node.inputs["Alpha"],
        )
        _ = links.new(
            input=bsdf_node.outputs["BSDF"],
            output=output_node.inputs["Surface"],
        )



from blenderline.collections.base import BaseCollection
from blenderline.entries.hdr import HDREntry


class HDRCollection(BaseCollection[HDREntry]):
    """Collection of registered HDR background entries."""

    pass



from blenderline.collections.base import BaseCollection
from blenderline.entries.item import ItemEntry


class ItemCollection(BaseCollection[ItemEntry]):
    """Collection of registered item entries."""

    pass



import bpy

from blenderline.collections import HDRCollection


class HDRManager:
    """Manager for HDR-related operations, such as creating a world texture and sampling
    and setting a random HDR background.
    """

    def __init__(self, hdr_collection: HDRCollection) -> None:
        """Create HDR background manager.

        Args:
            hdr_collection (HDRCollection): collection of HDR background entries.
        """
        # Save object attributes.
        self.hdr_collection = hdr_collection

    def initialize(self) -> None:
        """Create world texture with nodes to put HDR background texture on."""
        # Create new world texture and enable nodes.
        self.world = bpy.data.worlds.new("world")
        self.world.use_nodes = True

        # Set new world as active world in scene.
        bpy.context.scene.world = self.world

    def sample(self) -> None:
        """Sample HDR background entry from collection and apply it."""
        # Sample HDR entry from collection and apply it to initialized world texture.
        hdr_entry = self.hdr_collection.sample()
        hdr_entry.set(self.world)



import random

import bpy
import mathutils


class PathReference:
    """Reference to path object in scene."""

    def __init__(self, path_object: bpy.types.Object) -> None:
        """Create reference to path object in scene.

        Args:
            path_object (bpy.types.Object): path object in scene.
        """
        # Save object attributes
        self.path_object = path_object

    def sample(self) -> tuple[mathutils.Vector, mathutils.Vector, mathutils.Vector]:
        """Sample random location along the path and return location, normal vector in
        the x-y plane, and a normal vector in the local z plane.
        """
        # Get curve data from path object. It is assumed that path object only has one
        # curve, which in turn only has one spline. No error will be thrown otherwise,
        # but unexpected locations may be sampled.
        path_curve: bpy.types.Curve = self.path_object.data
        bezier_points = path_curve.splines[0].bezier_points

        # Curves consist of segments, which are defined between two control points. Each
        # segment adds 1 to the domain of t, the parameter defining a location on the
        # curve. For example, t=1.5 denotes the halfway mark of the second segment,
        # whereas t=0.5 denotes the halfway mark of the first segment.
        segment_index, t = divmod(random.random() * (len(bezier_points) - 1), 1)
        segment_index = int(segment_index)

        # Each segment is a cubic bezier curve, on which a point can be computed using the
        # explicit formula shown on wikipedia.
        p0: mathutils.Vector = bezier_points[segment_index].co
        p1: mathutils.Vector = bezier_points[segment_index].handle_right
        p2: mathutils.Vector = bezier_points[segment_index + 1].handle_left
        p3: mathutils.Vector = bezier_points[segment_index + 1].co
        location = (
            (1 - t) ** 3 * p0
            + 3 * t * (1 - t) ** 2 * p1
            + 3 * t**2 * (1 - t) * p2
            + t**3 * p3
        )

        # Compute location tangent by computing derivative of point using the explicit
        # formula shown on wikipedia.
        tangent = (
            3 * (1 - t) ** 2 * (p1 - p0)
            + 6 * t * (1 - t) * (p2 - p1)
            + 3 * t**2 * (p3 - p2)
        )
        tangent.normalize()

        # Compute normal to location by taking cross product of tangent line and z-vector.
        # This produces a normal vector in the x-y plane, which is used for lateral
        # offsetting of items on the path.
        normal_global_xy = tangent.cross(mathutils.Vector((0, 0, -1)))
        normal_global_xy.normalize()

        # Take cross product of tangent line and normal in the x-y plane to get normal to
        # location in the local z direction, which is used for orientation of items on
        # sloped paths.
        normal_local_z = tangent.cross(normal_global_xy)
        normal_local_z.normalize()

        return location, normal_global_xy, normal_local_z



import pathlib
import secrets

import bpy

from blenderline.references import ItemReference


class SceneManager:
    """Manager for scene-related operations, such as loading a scene and configuring the
    camera.
    """

    def __init__(
        self,
        filepath: pathlib.Path,
        camera_object_name: str,
        render_samples: int,
        render_use_cuda: bool,
        render_denoising: bool,
        render_resolution: list[int, int],
    ) -> None:
        """Create scene manager.

        Args:
            filepath (pathlib.Path): absolute filepath to scene .blend asset.
            camera_object_name (str): name of camera object in scene.
            render_samples (int): number of samples to render with.
            render_use_cuda (bool): whether to use CPU (False) or CUDA GPU (True).
            render_denoising (bool): enable denoising on rendered image.
            render_resolution (list[int, int]): image resolution ([x, y]) to render at.
        """
        # Save object attributes
        self.filepath = filepath
        self.camera_object_name = camera_object_name
        self.render_samples = render_samples
        self.render_use_cuda = render_use_cuda
        self.render_denoising = render_denoising
        self.render_resolution = render_resolution

    def initialize(self) -> None:
        """Initialize scene using specified parameters by loading scene and configuring
        the camera.
        """
        self.load_scene()
        self.configure_camera()

    def load_scene(self) -> None:
        """Load scene .blend file as main Blender file."""
        # Start with empty Blender file.
        bpy.ops.wm.read_factory_settings(use_empty=True)

        # Add collection to copy all scene objects to from scene .blend file.
        scene_collection = bpy.data.collections.new("scene")
        bpy.context.scene.collection.children.link(scene_collection)

        # Copy all objects from specified scene .blend file. Note that collection
        # organization of original scene .blend file is lost, as all objects are placed
        # in a new collection named "scene".
        with bpy.data.libraries.load(str(self.filepath)) as (data_from, data_to):
            data_to.objects = data_from.objects

        for object in data_to.objects:
            if object is not None:
                scene_collection.objects.link(object)

    def configure_camera(self) -> None:
        """Configure active camera in scene."""
        # Get camera object and set it as active in the scene.
        camera_object = bpy.data.objects[self.camera_object_name]
        bpy.context.scene.camera = camera_object

        # Set render engine to cycles, as eevee does not support headless rendering.
        bpy.context.scene.render.engine = "CYCLES"

        # Enable CUDA GPU
        if self.render_use_cuda:
            cycles_preferences = bpy.context.preferences.addons["cycles"].preferences
            cycles_preferences.compute_device_type = "CUDA"
            cycles_preferences.get_devices()
            bpy.context.scene.cycles.device = "GPU"

        # Configure render samples, denoising, and resolution.
        bpy.context.scene.cycles.samples = self.render_samples
        bpy.context.scene.cycles.use_denoising = self.render_denoising
        bpy.context.scene.render.resolution_x = self.render_resolution[0]
        bpy.context.scene.render.resolution_y = self.render_resolution[1]

    def reset_compositor_nodes(self) -> None:
        """Configure render settings."""
        # Enable object pass indexin view layer.
        bpy.context.scene.view_layers[0].use_pass_object_index = True

        # Enable nodes on scene compositor and get node tree.
        bpy.context.scene.use_nodes = True
        self.node_tree = bpy.context.scene.node_tree
        self.nodes = self.node_tree.nodes

        # Clear all nodes.
        self.nodes.clear()

        # Add Render Layers node.
        node: bpy.types.CompositorNodeRLayers = self.nodes.new("CompositorNodeRLayers")
        self.render_layers_node = node
        self.render_layers_node.location = (-300, 0)

        # Generate random indentifier for image
        image_filename = "image__" + secrets.token_hex(6) + "__"

        # Add File Output node. Save node as instance attribute to change output path.
        node: bpy.types.CompositorNodeOutputFile = self.nodes.new(
            "CompositorNodeOutputFile"
        )
        self.file_output_node = node
        self.file_output_node.file_slots.remove(node.inputs["Image"])
        self.file_output_node.file_slots.new(image_filename)
        self.file_output_node.format.color_mode = "RGB"
        self.file_output_node.location = (300, 0)

        # Link nodes.
        self.links = self.node_tree.links
        _ = self.links.new(
            input=self.render_layers_node.outputs["Image"],
            output=self.file_output_node.inputs[image_filename],
        )

    def add_item_reference_render_output(self, item_reference: ItemReference) -> None:
        """Add nodes to get segmentation mask corresponding to an item reference.

        Args:
            item_reference (ItemReference): item reference to generate mask output for.
        """
        # Generate object segmentation mask output filename
        mask_filename = (
            "mask__"
            + str(item_reference.reference_entry.label)
            + "__"
            + secrets.token_hex(6)
            + "__"
        )

        # Add output file for object segmentation mask to File Output node.
        self.file_output_node.file_slots.new(mask_filename)

        # Add ID Mask node.
        node: bpy.types.CompositorNodeIDMask = self.nodes.new("CompositorNodeIDMask")
        id_mask_node = node
        id_mask_node.index = item_reference.pass_index
        id_mask_node.use_antialiasing = True

        # Link nodes.
        _ = self.links.new(
            input=self.render_layers_node.outputs["IndexOB"],
            output=id_mask_node.inputs["ID value"],
        )
        _ = self.links.new(
            input=id_mask_node.outputs["Alpha"],
            output=self.file_output_node.inputs[mask_filename],
        )

    def render(
        self,
        output_folder: pathlib.Path,
        item_references: list[ItemReference],
    ) -> None:
        """Render current scene, outputting rendered image and all item segmentation
            masks. Rendered image will have filename "image__0001.png". Segmentation masks
            will have filename "<label ID>__<random item ID>__0001.png".

        Args:
            output_folder (pathlib.Path): folder to store images in.
            item_references (list[ItemReference]): list of item refereces to generate
                masks for.
        """
        # Reset compositor nodes.
        self.reset_compositor_nodes()

        # Set output folder.
        self.file_output_node.base_path = str(output_folder)

        # Add item references to segmentation mask outputs.
        for item_reference in item_references:
            self.add_item_reference_render_output(item_reference)

        # Start render.
        bpy.ops.render.render()



import random
from typing import Generic, TypeVar

from blenderline.entries.base import BaseEntry

EntryType = TypeVar("EntryType", bound=BaseEntry)


class BaseCollection(Generic[EntryType]):
    """Base class for collections."""

    def __init__(self) -> None:
        """Initialize empty collection."""
        self.entries: list[EntryType] = []

    def register(self, entry: EntryType) -> None:
        """Register new entry to collection.

        Args:
            entry (EntryType): entry to register to the collection.
        """
        self.entries.append(entry)

    @property
    def total_frequency(self) -> float:
        """Total frequency of all entries in the collection.

        Returns:
            float: total frequency.
        """
        return sum(entry.relative_frequency for entry in self.entries)

    def sample(self) -> EntryType:
        """Sample from registered entries according to computed probabilities.

        Returns:
            EntryType: sampled entry.
        """
        # Compute individual entry sampling probabilities and sample accordingly.
        weights = [
            entry.relative_frequency / self.total_frequency for entry in self.entries
        ]
        return random.choices(self.entries, weights=weights, k=1)[0]



import json
import pathlib
from dataclasses import dataclass

from blenderline.managers import (
    BackgroundManager,
    HDRManager,
    ItemManager,
    SceneManager,
)


@dataclass(frozen=True, eq=True)
class Split:
    """Split in dataset."""

    name: str
    size: int


@dataclass(frozen=True, eq=True)
class Label:
    """Label in dataset"""

    label: int
    label_name: str


class ImageDatasetGenerator:
    """Generator for image datasets."""

    def __init__(
        self,
        name: str,
        target: pathlib.Path,
        scene_manager: SceneManager,
        hdr_manager: HDRManager,
        background_manager: BackgroundManager,
        item_manager: ItemManager,
    ) -> None:
        """Create dataset generator.

        Args:
            name (str): name of dataset. Output folder will be target/<name>/<split>.
            target (pathlib.Path): base directory in which to put data.
            scene_manager (SceneManager): scene manager.
            hdr_manager (HDRManager): HDR background manager.
            background_manager (BackgroundManager): background manager.
            item_manager (ItemManager): item manager.
        """
        # Save object attributes.
        self.name = name
        self.target = target
        self.scene_manager = scene_manager
        self.hdr_manager = hdr_manager
        self.background_manager = background_manager
        self.item_manager = item_manager

        # Keep track of registered splits and labels.
        self.registered_splits: list[Split] = list()
        self.registered_labels: set[Label] = set()

    def register_split(self, name: str, size: int) -> None:
        """Register dataset split to generate

        Args:
            name (str): name of split to generate within dataset name output folder.
            size (int): number of images to generate.
        """
        self.registered_splits.append(Split(name, size))

    def register_label(self, label: int, label_name: str) -> None:
        """Register dataset label to generate

        Args:
            label (int): label index found in mask filenames.
            label_name (str): semantic name for label index.
        """
        self.registered_labels.add(Label(label, label_name))

    def initialize(self) -> None:
        self.scene_manager.initialize()
        self.hdr_manager.initialize()
        self.background_manager.initialize()
        self.item_manager.initialize()

    def generate_dataset(self) -> None:
        for split in self.registered_splits:
            for i in range(split.size):
                # Randomly sample (HDR) background.
                self.hdr_manager.sample()
                self.background_manager.sample()

                # Sample number of items and assign pass indices
                self.item_manager.sample()
                self.item_manager.assign_pass_indices()

                # Render image and segmentation masks
                self.scene_manager.render(
                    output_folder=self.target / self.name / split.name / str(i),
                    item_references=self.item_manager.item_references,
                )

                # Clear items for next iteration
                self.item_manager.clear()

        # Write label mapping
        with open(self.target / self.name / "label_mapping.json", mode="+wt") as file:
            label_mapping = {
                label.label: label.label_name for label in self.registered_labels
            }
            json.dump(label_mapping, file)



import json
import os
import pathlib
import re
from dataclasses import dataclass


@dataclass(frozen=True, eq=True)
class BlenderLineImage:
    """Reference to generated image within instance folder."""

    id: str
    path: pathlib.Path


@dataclass(frozen=True, eq=True)
class BlenderLineMask:
    """Reference to generated mask within instance folder."""

    id: str
    path: pathlib.Path
    label: str  # Label index (string, as label indices in label_mapping.json are strings)


def get_blenderline_image(instance_path: pathlib.Path) -> BlenderLineImage:
    """Get reference to generated image in a BlenderLine instance folder.

    Args:
        instance_path (pathlib.Path): path to BlenderLine instance folder.

    Returns:
        BlenderLineImage: reference to image containing path and image ID.
    """
    # BlenderLine names images with format "image__<image ID>__0001.png", where the image
    # ID is a 6 byte random hex string.
    image_pattern = re.compile(r"image__([0-9a-f]{12})__0001.png")

    matches: list[BlenderLineImage] = []

    for path in instance_path.iterdir():
        if match := image_pattern.search(path.name):
            matches.append(BlenderLineImage(id=match.group(1), path=path))

    # Check that the instance contains exactly one generated image.
    if len(matches) != 1:
        raise Exception(
            f"Instance at {instance_path} does not contain a valid BlenderLine image."
        )

    return matches[0]


def get_blenderline_masks(instance_path: pathlib.Path) -> list[BlenderLineMask]:
    """Get references to generated masks in a BlenderLine instance folder.

    Args:
        instance_path (pathlib.Path): path to BlenderLine instance folder.

    Returns:
        list[BlenderLineMask]: references to masks containing path, mask ID, and label ID.
    """
    # BlenderLine names masks with format "mask__<label ID>__<image ID>__0001.png", where
    # the mask ID is a 6 byte random hex string, and the label ID must be a digit.
    # TODO: there is currently no check for label IDs that are not digits, which are
    # incompatible with the YOLO metadata format.
    mask_pattern = re.compile(r"mask__([0-9]{1,10})__([0-9a-f]{12})__0001.png")

    matches: list[BlenderLineMask] = []

    for path in instance_path.iterdir():
        if match := mask_pattern.search(path.name):
            matches.append(
                BlenderLineMask(id=match.group(2), path=path, label=match.group(1))
            )

    return matches


def get_label_mapping(source_path: pathlib.Path) -> dict[str, str]:
    """Get label mapping generated by BlenderLine.

    Args:
        source_path (pathlib.Path): path to BlenderLine generated dataset root folder.

    Returns:
        dict[str, str]: mappign between label IDs and human-readable label names.
    """
    # Check for label mapping file existence.
    if not "label_mapping.json" in os.listdir(source_path):
        raise Exception(
            f"Source folder at {source_path} does not contain a label mapping."
        )

    with open(source_path / "label_mapping.json") as file:
        label_mapping = json.load(file)

    return label_mapping



import argparse
import pathlib
import sys

blenderline_dir = pathlib.Path(__file__).parent.parent
sys.path.append(str(blenderline_dir))

from blenderline.scripts.python import run_convert, run_generate  # noqa: E402


def cli_parser() -> argparse.ArgumentParser:
    # Main parser
    parser = argparse.ArgumentParser(
        prog="blenderline",
        description="BlenderLine: a pipeline for generating synthetic production line images.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        dest="command", help="Select a BlenderLine command to run:"
    )

    # Generate subparser
    generate_parser = subparsers.add_parser(
        name="generate",
        help="Generate a BlenderLine dataset from a configuration file.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    generate_required_parser = generate_parser.add_argument_group("required arguments")
    generate_required_parser.add_argument(
        "--config",
        required=True,
        metavar="<filepath>",
        help="Absolute or relative location of the configuration file.",
    )
    generate_optional_parser = generate_parser.add_argument_group("optional arguments")
    generate_optional_parser.add_argument(
        "--target",
        required=False,
        metavar="<filepath>",
        help="Absolute or relative location of the generated dataset.\n"
        "By default, BlenderLine generates the dataset in the current working directory.",
    )
    generate_optional_parser.add_argument(
        "--blender",
        required=False,
        metavar="<filepath>",
        help="Absolute location of the folder in which Blender is installed.\n"
        "By default, BlenderLine assumes that Blender is added to the system path.",
    )

    # Convert subparser
    convert_parser = subparsers.add_parser(
        name="convert",
        help="Convert a BlenderLine dataset to another format, e.g., yolo.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    convert_required_parser = convert_parser.add_argument_group("required arguments")
    format_choices = ["yolo_detection"]
    convert_required_parser.add_argument(
        "--format",
        required=True,
        choices=format_choices,
        metavar="<option>",
        help=f"Type of dataset to convert to. Must be in {{{', '.join(format_choices)}}}",
    )
    convert_required_parser.add_argument(
        "--source",
        required=True,
        metavar="<filepath>",
        help="Absolute or relative location of the BlenderLine dataset to convert.",
    )
    convert_optional_parser = convert_parser.add_argument_group("optional arguments")
    convert_optional_parser.add_argument(
        "--target",
        required=True,
        metavar="<filepath>",
        help="Absolute or relative location of the converted dataset.",
    )
    convert_optional_parser.add_argument(
        "--minarea",
        required=False,
        default=0.005,
        metavar="<float>",
        help="Minimum area an object mask must have to be included in the dataset.\n"
        "Area is defined as (num_pixels_object_mask/num_pixels_total_image).\n"
        "By default, BlenderLine omits masks with an area less than 0.005.",
    )
    convert_flags_parser = convert_parser.add_argument_group("flags")
    convert_flags_parser.add_argument(
        "--remove",
        required=False,
        action="store_true",
        help="If set, the source dataset is removed after conversion. While technically\n"
        "possible, it is discouraged to add the converted dataset to the source dataset\n"
        "by setting neither --target nor --remove.",
    )

    return parser


def cli():
    parser = cli_parser()

    args = parser.parse_args()

    if args.command == "generate":
        run_generate(config=args.config, target=args.target, blender=args.blender)
    elif args.command == "convert":
        run_convert(
            format=args.format,
            source=args.source,
            target=args.target,
            minarea=args.minarea,
            remove=args.remove,
        )


if __name__ == "__main__":
    cli()



import argparse
import pathlib
import sys

# Add base dir to PATH for module discovery within Blender
blenderline_dir = pathlib.Path(__file__).parent.parent.parent.parent
sys.path.append(str(blenderline_dir))

from blenderline.settings import ImageDatasetSettings  # noqa: E402


def main() -> None:
    # Build parser for arguments supplied to blender after "--"
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        required=True,
        metavar="<filepath>",
        help="Absolute location of the configuration file.",
    )
    parser.add_argument(
        "--target",
        required=True,
        metavar="<filepath>",
        help="Absolute location of the directory where the dataset is generated.",
    )

    # Parse arguments after "--".
    if "--" not in sys.argv:
        args = parser.parse_args([])
    else:
        args = parser.parse_args(sys.argv[sys.argv.index("--") + 1 :])

    # Get settings object.
    image_dataset_settings = ImageDatasetSettings(
        config=pathlib.Path(args.config),
        target=pathlib.Path(args.target),
    )

    # Generate dataset.
    image_dataset_generator = image_dataset_settings.get_dataset_generator()
    image_dataset_generator.initialize()
    image_dataset_generator.generate_dataset()


if __name__ == "__main__":
    main()



import os
import pathlib

from blenderline.scripts.python.converters import run_convert_yolo_detection

CONVERTERS = {
    "yolo_detection": run_convert_yolo_detection,
}


def run_convert(
    format: str, source: str, target: str, minarea: float, remove: bool
) -> None:
    # Get absolute path to source folder and check that it is valid, i.e., exists and
    # contains a label_mapping.json file.
    source_path = pathlib.Path(os.path.abspath(source))
    if not source_path.is_dir() or "label_mapping.json" not in os.listdir(source_path):
        raise Exception(
            "Please specify a valid source directory in BlenderLine format."
        )

    # Get absolute path to target folder and check that it is valid, i.e., does not exist
    # or is empty. Also check if the target directory is not equal to the source directory
    # This is probably not needed however, as the source directory is likely not empty.
    target_path = pathlib.Path(os.path.abspath(target))
    if target_path == source_path:
        raise Exception("Please choose a different target directory than the source.")
    if target_path.exists() and any(target_path.iterdir()):
        raise Exception("Please make sure that the target directory is empty.")

    target_path.mkdir(parents=True, exist_ok=True)

    # Run converter corresponding to specified target format.
    convertor = CONVERTERS[format]
    convertor(source_path, target_path, minarea, remove)



import os
import pathlib
import signal
import subprocess
import sys

blenderline_dir = pathlib.Path(__file__).parent.parent.parent.parent
sys.path.append(str(blenderline_dir))


def run_generate(config: str, target: str, blender: str) -> None:
    # Get absolute path to configuration file and check that it is valid, i.e., exists
    # and is a JSON file.
    config_path = pathlib.Path(os.path.abspath(config))
    if not config_path.is_file() or config_path.suffix != ".json":
        raise Exception("Please specify a valid configuration file.")

    # Get absolute path to target directory if given, else, set target directory to
    # current working directory. Existence need not be checked, as Blender creates
    # all intermediate folders in the rendering process.
    if target:
        target_path = pathlib.Path(os.path.abspath(target))
    else:
        target_path = pathlib.Path(os.getcwd())

    # Append Blender start command to path if given, else, assume Blender start command
    # is available on path and can be called using `blender`. No verification is performed
    # on the given path, as this is likely platform-dependent.
    if blender:
        blender_path = pathlib.Path(blender) / "blender"
    else:
        blender_path = "blender"

    # Get absolute path to generate script to be executing in Blender context.
    script_path = (
        blenderline_dir / "blenderline" / "scripts" / "blender" / "generate.py"
    )

    # Start Blender process.
    blender_process = subprocess.Popen(
        [
            str(blender_path),
            "--background",
            "--python",
            str(script_path),
            "--",
            "--config",
            str(config_path),
            "--target",
            str(target_path),
        ]
    )

    # Handle termination of the subprocess
    signal.signal(signal.SIGTERM, lambda _signum, _frame: blender_process.terminate())

    try:
        blender_process.wait()
    except KeyboardInterrupt:
        try:
            blender_process.terminate()
        except OSError:
            pass
        blender_process.wait()

    sys.exit(blender_process.returncode)



import math
import random

import bpy
import mathutils

from blenderline.collections import ItemCollection
from blenderline.references import ItemReference, PathReference


class ItemManager:
    """Manager for item-related operations, such as spawning multiple items randomly
    along the defined path object.
    """

    def __init__(
        self,
        path_object_name: str,
        spawn_probability: float,
        max_tries: int,
        max_items: int,
        item_collection: ItemCollection,
    ) -> None:
        """Create item manager.

        Args:
            path_object_name (str): name of path object in scene.
            spawn_probability (float): probability to spawn a new item.
            max_tries (int): maximum number of times to attempt to spawn an item. Spawning
                may fail due to item minimum margin distance collisions.
            max_items (int): maximum number of items to spawn.
            item_collection (ItemCollection): collection of item entries.
        """
        # Save object attributes
        self.path_object_name = path_object_name
        self.spawn_probability = spawn_probability
        self.max_tries = max_tries
        self.max_items = max_items
        self.item_collection = item_collection
        self.item_references: list[ItemReference] = []

    def initialize(self) -> None:
        """Get path reference object."""
        # Get path object by name and wrap it in a path reference object.
        path_object = bpy.data.objects[self.path_object_name]
        self.path_reference = PathReference(path_object)

        # Create collection in which spawned items are placed for easy tracking
        self.scene_item_collection = bpy.data.collections.new("items")
        bpy.context.scene.collection.children.link(self.scene_item_collection)

    def location_is_valid(
        self,
        proposed_location: mathutils.Vector,
        proposed_min_margin_distance: float,
    ) -> bool:
        """Check if proposed object location is valid with respect to all currently
            spawned items.

        Args:
            proposed_location (mathutils.Vector): proposed location for new item.
            proposed_min_margin_distance (float): minimum margin distance for proposed item.

        Returns:
            bool: True if proposed location is valid w.r.t. all spawned items.
        """
        # Return False if any spawned object returns not valid.
        for item_reference in self.item_references:
            if not item_reference.location_is_valid(
                proposed_location, proposed_min_margin_distance
            ):
                return False

        return True

    def max_items_reached(self) -> bool:
        """Checks if another item may be spawned according to maximum number of items
        configured.
        """
        num_spawned_items = len(self.item_references)
        return num_spawned_items >= self.max_items

    def trial_success(self) -> bool:
        """Perform random draw according to success probability based on number of items
        currently spawned."""
        num_spawned_items = len(self.item_references)
        success_probability = math.exp((self.spawn_probability - 1) * num_spawned_items)
        return random.random() < success_probability

    def sample(self) -> None:
        """Sample a set of items along the path. Items are sampled until either the
        maximum number of items is reached, the random trial fails, or an item cannot
        be spawned within a maximum number of tries.
        """
        # Keep track of amount of tries to
        tries = 0

        # Keep sampling until maximum items have been reached or random trial fails.
        while (
            not self.max_items_reached()
            and self.trial_success()
            and tries < self.max_tries
        ):
            # Sample item entry from the collection to spawn in.
            item_entry = self.item_collection.sample()

            # Sample location along the path and compute normals.
            location, normal_lateral, normal_upright = self.path_reference.sample()

            # Randomly offset item laterally by sampling random number in [-1, 1] and
            # multiplying by the item's maximum lateral offset. Multiplying this offset
            # distance with the lateral normal vector determines the change to the
            # initially sampled location.
            offset_fraction = random.random() * 2 - 1
            offset_distance = offset_fraction * item_entry.max_lateral_distance
            lateral_offset_vector = normal_lateral * offset_distance
            location += lateral_offset_vector

            # Check validity of sampled location and place item if valid.
            if self.location_is_valid(location, item_entry.max_lateral_distance):
                # Spawn item and save reference.
                item_reference = item_entry.spawn(location)
                self.item_references.append(item_reference)

                # Orient item to path upright normal.
                item_reference.orient_to_vector(normal_upright)

                # Set tries to 0 if an item could be spawned succesfully.
                tries = 0
            else:
                # Increment counter if sampled location is not valid.
                tries += 1

    def assign_pass_indices(self) -> None:
        """Assign pass indices to all currently spawned items."""
        # Set pass index on every spawned item.
        for pass_index, item_reference in enumerate(self.item_references):
            item_reference.set_pass_index(pass_index + 1)  # Add one as background is 0.

    def clear(self) -> None:
        """Remove all currently spawned item references from the scene."""
        while self.item_references:
            self.item_references.pop().delete()



import json
import pathlib

from blenderline.collections import BackgroundCollection, HDRCollection, ItemCollection
from blenderline.entries import BackgroundEntry, HDREntry, ItemEntry
from blenderline.generators import ImageDatasetGenerator
from blenderline.managers import (
    BackgroundManager,
    HDRManager,
    ItemManager,
    SceneManager,
)


class ImageDatasetSettings:
    """TODO"""

    def __init__(self, config: pathlib.Path, target: pathlib.Path) -> None:
        """TODO

        Args:
            config (pathlib.Path): Absolute location of the configuration file.
            target (pathlib.Path): Absolute location of the directory where the dataset
                is generated.
        """
        # Save base directory and load JSON config object.
        self.target = target
        with open(config, mode="rt") as file:
            self.settings: dict = json.load(file)

        # Get configuration file directory.
        self.config_dir = config.parent

    def get(self, key: str, default=None):
        """Get nested value from settings dictionary using dot notation to specify
            a path of keys to follow. E.g., "scene.path". Intermediate keys will return
            an empty dictionary of not present, so the function will always follow the
            entire key path.

        Args:
            key (str): dot-separated path of keys to follow
            default (_type_, optional): default value to return if key does not exist.
                Note that this is also triggered if an intermediary key is not present.
                Defaults to None.
        """
        parts = key.split(".")

        # If only one key, perform normal dictionary get on settings dictionary.
        if len(parts) == 1:
            return self.settings.get(parts[0], default)
        # If two keys, get first key first with empty dictionary as default.
        elif len(parts) == 2:
            return self.settings.get(parts[0], {}).get(parts[1], default)
        # For more than three keys (n), get the first n-1 keys with an empty dictionary
        # as default value.
        else:
            current_dict = self.settings
            for part in parts[:-1]:
                current_dict = current_dict.get(part, {})
            return current_dict.get(parts[-1], default)

    def get_hdr_collection(self) -> HDRCollection:
        """Create HDR collection from registered HDR backgrounds in settings.

        Returns:
            HDRCollection: collection of HDR backgrounds.
        """
        hdr_collection = HDRCollection()

        # Get list of registered HDR dictionaries.
        hdrs: list[dict] = self.get("hdrs.entries", [])

        for hdr in hdrs:
            # Validate registered HDR dict by checking for a relative filepath.
            if "path" not in hdr:
                raise Exception("Configure path to HDR asset")

            # Build registered HDR object from dict and register it to the collection.
            hdr_collection.register(
                HDREntry(
                    filepath=self.config_dir / hdr["path"],
                    relative_frequency=hdr.get("relative_frequency", 1),
                )
            )

        return hdr_collection

    def get_background_collection(self) -> BackgroundCollection:
        """Create background collection from registered backgrounds in settings.

        Returns:
            BackgroundCollection: collection of backgrounds.
        """
        background_collection = BackgroundCollection()

        # Get list of registered background dictionaries.
        backgrounds: list[dict] = self.get("backgrounds.entries", [])

        for background in backgrounds:
            # Validate registered HDR dict by checking for a relative filepath.
            if "path" not in background:
                raise Exception("Configure path to background asset")

            # Build registered HDR object from dict and register it to the collection.
            background_collection.register(
                BackgroundEntry(
                    filepath=self.config_dir / background["path"],
                    relative_frequency=background.get("relative_frequency", 1),
                )
            )

        return background_collection

    def get_item_collection(self) -> ItemCollection:
        """Create item collection from registered items in settings.

        Returns:
            ItemCollection: collection of items.
        """
        item_collection = ItemCollection()

        # Get list of registered item dictionaries.
        items: list[dict] = self.get("items.entries", [])

        for item in items:
            # Validate registered item dict by checking for a relative filepath, label,
            # object name and margin.
            if "path" not in item:
                raise Exception("Configure path to item asset")
            if "label" not in item:
                raise Exception("Configure item label")
            if "label_name" not in item:
                raise Exception("Configure item label name")
            if "object_name" not in item:
                raise Exception("Configure name of object in .blend file")
            if "min_margin_distance" not in item:
                raise Exception("Configure minimum distance to other items")
            if "max_lateral_distance" not in item:
                raise Exception("Configure maximum distance from path center line")

            # Build registered HDR object from dict and register it to the collection.
            item_collection.register(
                ItemEntry(
                    filepath=self.config_dir / item["path"],
                    label=item["label"],
                    object_name=item["object_name"],
                    min_margin_distance=item["min_margin_distance"],
                    max_lateral_distance=item["max_lateral_distance"],
                    relative_frequency=item.get("relative_frequency", 1),
                )
            )

        return item_collection

    def get_scene_manager(self) -> SceneManager:
        """Create scene manager using parameters configured in settings.

        Returns:
            SceneManager: scene manager object.
        """
        # Validate scene settings by checking for a relative filepath.
        if not self.get("scene.path"):
            raise Exception("Configure path to scene asset")

        # Return scene manager with specified parameters, falling back to defaults if not
        # specified.
        return SceneManager(
            filepath=self.config_dir / self.settings["scene"]["path"],
            camera_object_name=self.get("scene.camera_object_name", "camera"),
            render_samples=self.get("scene.render_samples", 1024),
            render_use_cuda=self.get("scene.render_use_cuda", False),
            render_denoising=self.get("scene.render_denoising", True),
            render_resolution=self.get("scene.render_resolution", [512, 512]),
        )

    def get_hdr_manager(self) -> HDRManager:
        """Create HDR background manager using parameters configured in settings.

        Returns:
            HDRManager: HDR manager object.
        """
        return HDRManager(hdr_collection=self.get_hdr_collection())

    def get_background_manager(self) -> BackgroundManager:
        """Create background manager using parameters configured in settings.

        Returns:
            BackgroundManager: background manager object.
        """
        return BackgroundManager(
            background_object_name=self.get(
                "scene.background_object_name", "background"
            ),
            background_collection=self.get_background_collection(),
        )

    def get_item_manager(self) -> ItemManager:
        """Create item manager using parameters configured in settings.

        Returns:
            ItemManager: item manager object.
        """
        return ItemManager(
            path_object_name=self.get("items.path_object_name", "path"),
            spawn_probability=self.get("items.spawn_probability", 0.5),
            max_tries=self.get("items.max_tries", 3),
            max_items=self.get("items.max_items", 5),
            item_collection=self.get_item_collection(),
        )

    def get_dataset_generator(self) -> ImageDatasetGenerator:
        """Create image dataset generator using parameters configured in settings.

        Returns:
            ImageDatasetGenerator: dataste generator object.
        """
        # Create object using settings.
        image_dataset_generator = ImageDatasetGenerator(
            name=self.get("dataset.name", "dataset"),
            target=self.target,
            scene_manager=self.get_scene_manager(),
            hdr_manager=self.get_hdr_manager(),
            background_manager=self.get_background_manager(),
            item_manager=self.get_item_manager(),
        )

        # Register all splits.
        for split_dict in self.get("dataset.splits", []):
            if "name" not in split_dict or "size" not in split_dict:
                raise Exception("Invalid split configured. Specify name and size keys")

            image_dataset_generator.register_split(
                name=split_dict["name"], size=split_dict["size"]
            )

        # Register all labels.
        for item_dict in self.get("items.entries", []):
            if "label" not in item_dict or "label_name" not in item_dict:
                raise Exception("Invalid item configured, Specify label and label name")

            # Build registered HDR object from dict and register it to the collection.
            image_dataset_generator.register_label(
                label=item_dict["label"], label_name=item_dict["label_name"]
            )

        return image_dataset_generator



import bpy

from blenderline.collections import BackgroundCollection


class BackgroundManager:
    """Manager for background-related operations, such as grabbing the background object
    and setting a random background.
    """

    def __init__(
        self, background_object_name: str, background_collection: BackgroundCollection
    ) -> None:
        """Create background manager.

        Args:
            background_object_name (str): name of background object in scene.
            background_collection (BackgroundCollection): collection of background entries.
        """
        # Save object attributes
        self.background_object_name = background_object_name
        self.background_collection = background_collection

    def initialize(self) -> None:
        """Get background object."""
        # Get background object by name.
        self.background_object = bpy.data.objects[self.background_object_name]

    def sample(self) -> None:
        """Sample background entry from collection and apply it."""
        # Sample background entry from collection and apply it to background object
        background_entry = self.background_collection.sample()
        background_entry.set(self.background_object)



import pathlib
import shutil

import cv2
import numpy as np
import yaml

from .utils import (
    BlenderLineMask,
    get_blenderline_image,
    get_blenderline_masks,
    get_label_mapping,
)


def get_yolo_label(mask: BlenderLineMask, minarea: float) -> str | None:
    """Get YOLO bounding box label from BlenderLine generated pixel mask.

    Args:
        mask (BlenderLineMask): reference to mask with image ID, label ID, and image path.
        minarea (float): minimum area an object mask must have to be included.

    Returns:
        str | None: YOLO bounding box with class ID, or None if minarea is not exceeded.
    """
    # Read grayscale mask and binarize.
    mask_gray = cv2.imread(str(mask.path), cv2.IMREAD_GRAYSCALE)
    _, mask_binary = cv2.threshold(mask_gray, 127, 1, cv2.THRESH_OTSU)

    # Return None if mask is too small (< area_threshold).
    mask_height, mask_width = mask_gray.shape
    if mask_binary.sum() / (mask_width * mask_height) < minarea:
        return None

    # Compute bounding box relative to mask width and height.
    rows, cols = np.where(mask_binary == 1)
    xmin, xmax = cols.min(), cols.max()
    ymin, ymax = rows.min(), rows.max()

    xcenter = (xmin + xmax) / 2 / mask_width
    ycenter = (ymin + ymax) / 2 / mask_height
    width = (xmax - xmin) / mask_width
    height = (ymax - ymin) / mask_height

    # Return line in YOLO dataset format.
    return f"{mask.label} {xcenter} {ycenter} {width} {height}"


def run_convert_yolo_detection(
    source_path: pathlib.Path, target_path: pathlib.Path, minarea: float, remove: bool
) -> None:
    """Convert BlenderLine generated dataset to YOLO object detection dataset format.

    Args:
        source_path (pathlib.Path): absolute location of dataset to convert.
        target_path (pathlib.Path): absolute location of converted dataset.
        minarea (float): minimum area an object mask must have to be included.
        remove (bool): whether to remove the source dataset after converting.
    """
    # Create image and label subfolders.
    images_path = target_path / "images"
    labels_path = target_path / "labels"
    images_path.mkdir()
    labels_path.mkdir()

    # Load BlenderLine label mapping and prepopulate YOLO metadata. Paths to training,
    # validation, and testing data will be added when looping over the data splits.
    label_mapping = get_label_mapping(source_path)
    metadata = {
        "path": "/".join([".."] + list(target_path.parts[-2:])),
        "train": None,
        "val": None,
        "test": None,
        "names": label_mapping,
    }

    # Detect data splits from root BlenderLine dataset directory.
    split_paths = [path for path in source_path.iterdir() if path.is_dir()]

    for split_path in split_paths:
        # Create split folder within image and label directories
        images_split_path = images_path / split_path.name
        labels_split_path = labels_path / split_path.name
        images_split_path.mkdir()
        labels_split_path.mkdir()

        # Automatically detect split names as training, validation, or testing.
        # This may be done more cleanly, but YOLO expects slightly unusual names such
        # as "val" for the validation set, and putting restrictions on the split
        # naming possibilities is undesired. In most cases, this should work fine.
        if "train" in (split := split_path.name):
            metadata["train"] = f"images/{split}"
        if "val" in (split := split_path.name):
            metadata["val"] = f"images/{split}"
        if "test" in (split := split_path.name):
            metadata["test"] = f"images/{split}"

        instance_paths = [path for path in split_path.iterdir() if path.is_dir()]

        for instance_path in instance_paths:
            # Get generated image and copy it over to target dataset location.
            image = get_blenderline_image(instance_path)
            shutil.copy(image.path, images_split_path / f"{image.id}.png")

            # Get generated masks, convert them to YOLO bounding box format and merge
            # into one label file corresponding to image ID.
            masks = get_blenderline_masks(instance_path)
            labels = [
                label for mask in masks if (label := get_yolo_label(mask, minarea))
            ]
            with open(labels_split_path / f"{image.id}.txt", "w+") as file:
                file.write("\n".join(labels))

    # Create YOLO metadata file.
    with open(target_path / f"{source_path.name}.yaml", "w+") as file:
        yaml.dump(metadata, file, sort_keys=False)

    # Remove source dataset if desired.
    if remove:
        choice = input(f"Are you sure you want to remove source {source_path}? [y/N]: ")
        if choice == "y":
            shutil.rmtree(source_path)
        else:
            print("Source removal aborted.")



import pathlib
import secrets

import bpy

from blenderline.entries.base import BaseEntry
from blenderline.references.item import ItemReference


class ItemEntry(BaseEntry):
    """Item entry."""

    def __init__(
        self,
        filepath: pathlib.Path,
        label: int,
        object_name: str,
        min_margin_distance: float,
        max_lateral_distance: float,
        relative_frequency: float,
    ) -> None:
        """Create item entry.

        Args:
            filepath (pathlib.Path): absolute filepath to item .blend asset.
            label (int): label index of object.
            object_name (str): name of object in item .blend file.
            min_margin_distance (float): minimum distance (in Blender units) other items
                need to be away from this item. Distance is computed between item center
                points.
            max_lateral_distance (float): maximum distance item can move from path.
            relative_frequency (float): relative frequency with which to sample.
        """
        # Save object attributes
        self.filepath = filepath
        self.label = label
        self.object_name = object_name
        self.min_margin_distance = min_margin_distance
        self.max_lateral_distance = max_lateral_distance
        self.relative_frequency = relative_frequency

    def spawn(self, location: tuple[float, float, float]) -> ItemReference:
        """Instantiate item object in the scene at a given location.

        Args:
            location (tuple[float, float, float]): location to spawn object at.

        Returns:
            ItemReference: reference to spawned item object.
        """
        scene_item_collection = bpy.data.collections["items"]

        with bpy.data.libraries.load(str(self.filepath)) as (data_from, data_to):
            data_to.objects = [
                name for name in data_from.objects if name == self.object_name
            ]

        for object in data_to.objects:
            if object is not None:
                scene_item_collection.objects.link(object)

        # Generate random name for object in scene to prevent object name
        # collisions when multiple objects are added to the scene
        scene_object_name = secrets.token_hex(6)

        # Select spawned object and set proper name and location
        item_object = bpy.data.objects[self.object_name]
        item_object.name = scene_object_name
        item_object.location = location

        return ItemReference(item_object, self)



import pathlib

import bpy

from blenderline.entries.base import BaseEntry


class HDREntry(BaseEntry):
    """HDR entry."""

    def __init__(
        self,
        filepath: pathlib.Path,
        relative_frequency: float,
    ) -> None:
        """Create HDR background entry.

        Args:
            filepath (pathlib.Path): absolute filepath to HDR asset.
            relative_frequency (float): relative frequency with which to sample.
        """
        # Save object attributes.
        self.filepath = filepath
        self.relative_frequency = relative_frequency

    def set(self, world: bpy.types.World) -> None:
        """Set HDR as environment background.

        Args:
            world (bpy.types.World): world object to apply background to.

        Sources:
            - https://blender.stackexchange.com/a/209633
        """
        # Get environment node tree of world.
        node_tree = world.node_tree
        nodes = node_tree.nodes

        # Clear all nodes.
        nodes.clear()

        # Add Environment Texture node.
        node: bpy.types.ShaderNodeTexEnvironment = nodes.new("ShaderNodeTexEnvironment")
        environment_node = node
        environment_node.image = bpy.data.images.load(str(self.filepath))
        environment_node.location = (-300, 0)

        # Add Background node.
        node: bpy.types.ShaderNodeBackground = nodes.new("ShaderNodeBackground")
        background_node = node
        background_node.location = (0, 0)

        # Add Output node.
        output_node = nodes.new("ShaderNodeOutputWorld")
        output_node.location = (300, 0)

        # Link nodes.
        links = node_tree.links
        _ = links.new(
            input=environment_node.outputs["Color"],
            output=background_node.inputs["Color"],
        )
        _ = links.new(
            input=background_node.outputs["Background"],
            output=output_node.inputs["Surface"],
        )



import bpy
import mathutils


class ItemReference:
    """Reference to spawned item object in scene."""

    def __init__(
        self,
        item_object: bpy.types.Object,
        reference_entry,  # No type hint to prevent circular import
    ) -> None:
        """Create reference to item object in scene.

        Args:
            object_name (str): item object in scene.
            reference_entry (ItemEntry): entry used to create referenced object.
        """
        # Save object attributes.
        self.item_object = item_object
        self.reference_entry = reference_entry

        # Prepare object for rotation by saving current orientation (up on initialization)
        self.orientation = mathutils.Vector((0, 0, 1))
        self.item_object.rotation_mode = "QUATERNION"

        # Set default pass index of object
        self.pass_index = 0

    def set_pass_index(self, pass_index: int) -> None:
        """Set pass index of object.

        Args:
            pass_index (int): pass index to set on object. Must be between 0 and 255.
        """
        self.item_object.pass_index = pass_index
        self.pass_index = pass_index

    def orient_to_vector(self, desired_orientation: mathutils.Vector) -> None:
        """Rotate object so that orientation (object local z axis) aligns with desired
            direction.

        Args:
            desired_orientation (mathutils.Vector): desired local z orientation of object.
        """
        # Determine normalized rotation quaternion between current and desired orientation.
        rotation_quaternion = self.orientation.rotation_difference(desired_orientation)
        rotation_quaternion.normalize()

        # Apply rotation and update object orientation.
        self.item_object.rotation_quaternion = rotation_quaternion
        self.orientation = desired_orientation

    def location_is_valid(
        self,
        proposed_location: mathutils.Vector,
        proposed_min_margin_distance: float,
    ) -> bool:
        """Check if a proposed location is far enough away from object, i.e., distance
            between the current object location and the proposed location is above the
            specified minimum margin distance defined on the reference item entry.

        Args:
            proposed_location (mathutils.Vector): proposed location for new item.
            proposed_min_margin_distance (float): minimum margin distance for proposed item.

        Returns:
            bool: True if proposed location is valid w.r.t. this item.
        """
        # Get location and minimum margin distance of spawned object
        current_location: mathutils.Vector = self.item_object.location
        current_min_margin_distance = self.reference_entry.min_margin_distance

        # Compute distance between current spawned object and proposed location
        distance = (current_location - proposed_location).length

        # Check if distance satisfies both minimum margin distances
        return distance > max(current_min_margin_distance, proposed_min_margin_distance)

    def delete(self) -> None:
        """Remove item object scene."""
        bpy.data.objects.remove(self.item_object, do_unlink=True)



from setuptools import find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="blenderline",
    version="0.1.6",
    description="A pipeline for generating synthetic production line images",
    packages=find_packages(exclude=["examples*", "data*"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MaxvandenHoven/blenderline",
    author="Max van den Hoven",
    author_email="max.hoven@gmail.com",
    license="GPLv3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "blender-stubs>=3.12.27",
        "bpy>=3.5.0",
        "numpy",
        "opencv-python",
        "pyyaml",
    ],
    extras_require={
        "dev": [
            "black>=23.3.0",
            "isort>=5.12.0",
            "flake8>=6.0.0",
            "twine>=4.0.2",
            "Sphinx==6.2.1",
            "sphinx-autobuild==2021.3.14",
            "furo==2023.5.20",
            "sphinx-copybutton==0.5.2",
            "sphinx_design==0.4.1",
        ],
    },
    python_requires="==3.10.*",
    entry_points={
        "console_scripts": [
            "blenderline = blenderline.command_line:cli",
        ],
    },
)



from blenderline.collections.base import BaseCollection
from blenderline.entries.background import BackgroundEntry


class BackgroundCollection(BaseCollection[BackgroundEntry]):
    """Collection of registered background entries."""

    pass
